import struct

import numpy as np
from numba.core import utils
import ctypes

from .abstract import *
from .containers import *
from .functions import *
from .iterators import *
from .misc import *
from .npytypes import *
from .scalars import *
from .function_type import *

numpy_version = tuple(map(int, np.__version__.split('.')[:2]))

# Short names

pyobject = PyObject('pyobject')
ffi_forced_object = Opaque('ffi_forced_object')
ffi = Opaque('ffi')
none = NoneType('none')
ellipsis = EllipsisType('...')
Any = Phantom('any')
undefined = Undefined('undefined')
py2_string_type = Opaque('str')
unicode_type = UnicodeType('unicode_type')
string = unicode_type
unknown = Dummy('unknown')
npy_rng = NumPyRandomGeneratorType('rng')
npy_bitgen = NumPyRandomBitGeneratorType('bitgen')

# _undef_var is used to represent undefined variables in the type system.
_undef_var = UndefVar('_undef_var')

code_type = Opaque('code')
pyfunc_type = Opaque('pyfunc')

# No operation is defined on voidptr
# Can only pass it around
voidptr = RawPointer('void*')

# optional types
optional = Optional
deferred_type = DeferredType
slice2_type = SliceType('slice<a:b>', 2)
slice3_type = SliceType('slice<a:b:c>', 3)
void = none

# Need to ignore mypy errors because mypy cannot unify types for both
# the type systems even if they're logically mutually exclusive.
# mypy: ignore-errors


boolean = bool_ = Boolean('bool')
if numpy_version >= (2, 0):
    bool = bool_

byte = uint8 = Integer('uint8')
uint16 = Integer('uint16')
uint32 = Integer('uint32')
uint64 = Integer('uint64')

int8 = Integer('int8')
int16 = Integer('int16')
int32 = Integer('int32')
int64 = Integer('int64')
intp = int32 if utils.MACHINE_BITS == 32 else int64
uintp = uint32 if utils.MACHINE_BITS == 32 else uint64
intc = int32 if struct.calcsize('i') == 4 else int64
uintc = uint32 if struct.calcsize('I') == 4 else uint64
ssize_t = int32 if struct.calcsize('n') == 4 else int64
size_t = uint32 if struct.calcsize('N') == 4 else uint64

float32 = Float('float32')
float64 = Float('float64')
float16 = Float('float16')

complex64 = Complex('complex64', float32)
complex128 = Complex('complex128', float64)

range_iter32_type = RangeIteratorType(int32)
range_iter64_type = RangeIteratorType(int64)
unsigned_range_iter64_type = RangeIteratorType(uint64)
range_state32_type = RangeType(int32)
range_state64_type = RangeType(int64)
unsigned_range_state64_type = RangeType(uint64)

signed_domain = frozenset([int8, int16, int32, int64])
unsigned_domain = frozenset([uint8, uint16, uint32, uint64])
integer_domain = signed_domain | unsigned_domain
real_domain = frozenset([float32, float64])
complex_domain = frozenset([complex64, complex128])
number_domain = real_domain | integer_domain | complex_domain

# Integer Aliases
c_bool = py_bool = np_bool_ = boolean

c_uint8 = np_uint8 = uint8
c_uint16 = np_uint16 = uint16
c_uint32 = np_uint32 = uint32
c_uint64 = np_uint64 = uint64
c_uintp = np_uintp = uintp

c_int8 = np_int8 = int8
c_int16 = np_int16 = int16
c_int32 = np_int32 = int32
c_int64 = np_int64 = int64
c_intp = py_int = np_intp = intp

c_float16 = np_float16 = float16
c_float32 = np_float32 = float32
c_float64 = py_float = np_float64 = float64

np_complex64 = complex64
py_complex = np_complex128 = complex128

# Domain Aliases
py_signed_domain = np_signed_domain = signed_domain
np_unsigned_domain = unsigned_domain
py_integer_domain = np_integer_domain = integer_domain
py_real_domain = np_real_domain = real_domain
py_complex_domain = np_complex_domain = complex_domain
py_number_domain = np_number_domain = number_domain

# Aliases to NumPy type names

b1 = bool_
i1 = int8
i2 = int16
i4 = int32
i8 = int64
u1 = uint8
u2 = uint16
u4 = uint32
u8 = uint64

f2 = float16
f4 = float32
f8 = float64

c8 = complex64
c16 = complex128

np_float_ = float32
np_double = double = float64
if numpy_version < (2, 0):
    float_ = float32

_make_signed = lambda x: globals()["int%d" % (ctypes.sizeof(x) * 8)]
_make_unsigned = lambda x: globals()["uint%d" % (ctypes.sizeof(x) * 8)]

char = np_char = _make_signed(ctypes.c_char)
uchar = np_uchar = byte = _make_unsigned(ctypes.c_ubyte)
short = np_short = _make_signed(ctypes.c_short)
ushort = np_ushort = _make_unsigned(ctypes.c_ushort)
int_ = np_int_ = np_intp
uint = np_uint = np_uintp
intc = np_intc = _make_signed(ctypes.c_int) # C-compat int
uintc = np_uintc = _make_unsigned(ctypes.c_uint) # C-compat uint
long_ = np_long = _make_signed(ctypes.c_long)  # C-compat long
ulong = np_ulong = _make_unsigned(ctypes.c_ulong)  # C-compat ulong
longlong = np_longlong = _make_signed(ctypes.c_longlong)
ulonglong = np_ulonglong = _make_unsigned(ctypes.c_ulonglong)

# This is equivalent to NumPy's `np.dtype('U1').itemsize`,
# which is the size of a single Unicode character in bytes.

# We can't keep this as `ctypes.c_wchar` because its size 
# is platform-dependent (2 bytes on Windows, 4 bytes on Unix).
sizeof_unicode_char = ctypes.sizeof(ctypes.c_byte) * 4

all_str = '''
int8
int16
int32
int64
uint8
uint16
uint32
uint64
intp
uintp
intc
uintc
ssize_t
size_t
boolean
float32
float64
complex64
complex128
bool_
byte
char
uchar
short
ushort
int_
uint
long_
ulong
longlong
ulonglong
float_
double
void
none
b1
i1
i2
i4
i8
u1
u2
u4
u8
f4
f8
c8
c16
optional
ffi_forced_object
ffi
deferred_type
'''

__all__ = all_str.split()
if numpy_version >= (2, 0):
    __all__.remove('float_')
    __all__.append('bool')

from numba.np.types.datetime import NPDatetime, NPTimedelta



import abc
import functools
from collections.abc import Hashable as _CanHash, Mapping, Sequence as _Sequence
from typing import (
    Any,
    Final,
    Literal as _Literal,
    Protocol,
    TypeAlias,
    overload,
    type_check_only,
)
from typing_extensions import ClassVar, Generic, Never, Self, TypeVar, override

from numba.core.typeconv import castgraph
from numba.core.typing import context, templates
from .npytypes import Array

###

_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_T_contra = TypeVar("_T_contra", contravariant=True)
_T_co = TypeVar("_T_co", covariant=True)
_TypeT_co = TypeVar("_TypeT_co", bound=Type, default=Type, covariant=True)

@type_check_only
class _CanCastPythonValue(Protocol[_T_contra, _T_co]):
    def cast_python_value(self, value: _T_contra, /) -> _T_co: ...

# the step is `Any` instead of `Literal[1] | None` so we can use it as invariant `list`
# element type
_SpecSlice: TypeAlias = slice[None, None, Any]
_Layout: TypeAlias = _Literal["F", "C", "A"]

###

class _TypeMetaclass(abc.ABCMeta):
    def __init__(
        cls, name: str, bases: tuple[type, ...], orig_vars: dict[str, Any]
    ) -> None: ...
    def __call__(cls, *args: Any, **kwargs: Any) -> Any: ...

class Type(metaclass=_TypeMetaclass):
    mutable: ClassVar[bool] = False
    reflected: ClassVar[bool] = False

    name: Final[str]

    def __init__(self, name: str) -> None: ...

    #
    @overload  # () -> Signature
    def __call__(self, /) -> templates.Signature: ...
    @overload  # (Type, ...) -> Signature
    def __call__(self, arg0: Type, /, *args: Type) -> templates.Signature: ...
    @overload  # (_: object) -> self.cast_python_value(_)
    def __call__(self: _CanCastPythonValue[_T1, _T2], arg0: _T1, /) -> _T2: ...

    #
    def __getitem__(
        self, args: _SpecSlice | tuple[_SpecSlice, ...] | list[_SpecSlice]
    ) -> Array: ...

    #
    @property
    def key(self) -> _CanHash: ...
    @property
    def mangling_args(self) -> tuple[str, tuple[object, ...]]: ...
    @property
    def is_internal(self) -> bool: ...

    #
    def can_convert_to(
        self,
        typingctx: context.Context,
        other: Type,
    ) -> castgraph.Conversion | None: ...
    def can_convert_from(
        self,
        typingctx: context.Context,
        other: Type,
    ) -> castgraph.Conversion | None: ...
    def cast_python_value(self, args: Never) -> Any: ...

    #
    def is_precise(self) -> bool: ...
    def augment(self, other: Type) -> Self | None: ...
    def unify(self, typingctx: context.Context, other: Type) -> Type | None: ...
    def dump(self, tab: str = "") -> None: ...

class Dummy(Type): ...
class Hashable(Type): ...

class Number(Hashable):
    @overload
    def unify(self, typingctx: context.Context, other: Number) -> Number: ...
    @overload
    def unify(self, typingctx: context.Context, other: Type) -> Number | None: ...

class Callable(Type):
    @abc.abstractmethod
    def get_call_type(
        self, context: context.Context, args: tuple[Type, ...], kws: Mapping[str, Type]
    ) -> templates.Signature | None: ...
    @abc.abstractmethod
    def get_call_signatures(self) -> tuple[_Sequence[templates.Signature], bool]: ...
    @abc.abstractmethod
    def get_impl_key(self, sig: templates.Signature) -> _CanHash: ...

class DTypeSpec(Type):
    @property
    @abc.abstractmethod
    def dtype(self) -> Any: ...

class IterableType(Type):
    @property
    @abc.abstractmethod
    def iterator_type(self) -> Type: ...

class Sized(Type): ...

class ConstSized(Sized):
    @abc.abstractmethod
    def __len__(self) -> int: ...

class IteratorType(IterableType):
    def __init__(self, name: str, **kwargs: Never) -> None: ...
    @property
    @abc.abstractmethod
    def yield_type(self) -> Type: ...
    @property
    def iterator_type(self) -> Self: ...

class Container(Sized, IterableType): ...
class Sequence(Container): ...

class MutableSequence(Sequence):
    mutable: ClassVar[bool] = True

class ArrayCompatible(Type):
    array_priority: ClassVar[float] = 0.0

    @property
    @abc.abstractmethod
    def as_array(self) -> ArrayCompatible: ...

    #
    @functools.cached_property
    def ndim(self) -> int: ...
    @functools.cached_property
    def layout(self) -> _Layout: ...
    @functools.cached_property
    def dtype(self) -> Type: ...

class Literal(Type, Generic[_T_co]):
    ctor_map: ClassVar[dict[type, type[Literal[Any]]]] = ...

    def __init__(self, value: _T_co) -> None: ...
    @property
    def literal_value(self) -> _T_co: ...
    @property
    def literal_type(self) -> Type: ...

class TypeRef(Dummy, Generic[_TypeT_co]):
    instance_type: _TypeT_co  # readonly

    def __init__(self, instance_type: _TypeT_co) -> None: ...
    @property
    @override
    def key(self) -> _TypeT_co: ...

class InitialValue(Generic[_T_co]):
    def __init__(self, initial_value: _T_co) -> None: ...
    @property
    def initial_value(self) -> _T_co: ...

class Poison(Type, Generic[_TypeT_co]):  # like `Never`
    def __init__(self, ty: _TypeT_co) -> None: ...
    def __unliteral__(self) -> Poison[Self]: ...
    @override
    def unify(self, typingctx: context.Context, other: Type) -> None: ...



"""
Memory monitoring utilities for measuring memory usage.

Example usage:
    tracker = MemoryTracker("my_function")
    with tracker.monitor():
        my_function()
    # Access data: tracker.rss_delta, tracker.duration, etc.
    # Get formatted string: tracker.get_summary()
"""
from __future__ import annotations
import os
import contextlib
import time
from typing import Dict, Optional


try:
    import psutil
    _HAS_PSUTIL = True
except ImportError:
    _HAS_PSUTIL = False


IS_SUPPORTED = _HAS_PSUTIL


def get_available_memory() -> Optional[int]:
    """
    Get current available system memory in bytes.

    Used for memory threshold checking in parallel test execution.

    Returns:
        int or None: Available memory in bytes, or None if unavailable
    """
    if _HAS_PSUTIL:
        try:
            sys_mem = psutil.virtual_memory()
            return sys_mem.available
        except Exception:
            pass

    return None


def get_memory_usage() -> Dict[str, Optional[int]]:
    """
    Get memory usage information needed for monitoring.

    Returns only RSS and available memory which are the fields
    actually used by the MemoryTracker.

    Returns:
        dict: Memory usage information including:
            - rss: Current process RSS (physical memory currently used)
            - available: Available system memory
    """
    memory_info = {}

    if _HAS_PSUTIL:
        try:
            # Get current process RSS
            process = psutil.Process(os.getpid())
            mem_info = process.memory_info()
            memory_info["rss"] = mem_info.rss

            # Get system available memory
            sys_mem = psutil.virtual_memory()
            memory_info["available"] = sys_mem.available

        except (psutil.NoSuchProcess, psutil.AccessDenied):
            pass

    # Set defaults if unavailable
    if "rss" not in memory_info:
        memory_info["rss"] = None
    if "available" not in memory_info:
        memory_info["available"] = None

    return memory_info


class MemoryTracker:
    """
    A simple memory monitor that tracks RSS delta and timing.

    Stores monitoring data in instance attributes for later access.
    Each instance is typically used for monitoring a single operation.
    """
    pid: int
    name: str
    start_time: float | None
    end_time: float | None
    start_memory: Dict[str, int | None] | None
    end_memory: Dict[str, int | None] | None
    duration: float | None
    rss_delta: int | None

    def __init__(self, name: str):
        """Initialize a MemoryTracker with empty monitoring data."""
        self.pid = os.getpid()
        self.name = name
        self.start_time = None
        self.end_time = None
        self.start_memory = None
        self.end_memory = None
        self.duration = None
        self.rss_delta = None

    @contextlib.contextmanager
    def monitor(self):
        """
        Context manager to monitor memory usage during function execution.

        Records start/end memory usage and timing, calculates RSS delta,
        and stores all data in instance attributes.

        Args:
            name (str): Name/identifier for the function or operation being
                        monitored

        Yields:
            self: The MemoryTracker instance for accessing stored data
        """
        # Store data in self and record start time and memory usage
        self.start_time = time.time()
        self.start_memory = get_memory_usage()

        try:
            yield self
        finally:
            # Record end time and memory usage
            self.end_time = time.time()
            self.end_memory = get_memory_usage()
            self.duration = self.end_time - self.start_time

            # Calculate RSS delta
            start_rss = self.start_memory.get("rss", 0)
            end_rss = self.end_memory.get("rss", 0)
            self.rss_delta = ((end_rss - start_rss)
                              if start_rss and end_rss else 0)

    def get_summary(self) -> str:
        """
        Return a formatted summary of the memory monitoring data.

        Formats the stored monitoring data into a human-readable string
        containing name, PID, RSS delta, available memory, duration,
        and start time.

        Returns:
            str: Formatted summary string with monitoring results

        Note:
            Should be called after monitor() context has completed
            to ensure all data is available.
        """
        if self.start_memory is None or self.end_memory is None:
            raise ValueError("Memory monitoring data not available")

        current_available = self.end_memory.get("available")

        def format_bytes(bytes_val, show_sign=False):
            """Convert bytes to human readable format"""
            if bytes_val is None:
                return "N/A"
            if bytes_val == 0:
                return "0 B"

            sign = ""
            if show_sign:
                sign = "-" if bytes_val < 0 else "+"
            bytes_val = abs(bytes_val)

            for unit in ["B", "KB", "MB", "GB"]:
                if bytes_val < 1024.0:
                    return f"{sign}{bytes_val:.2f} {unit}"
                bytes_val /= 1024.0
            return f"{sign}{bytes_val:.2f} TB"

        start_ts = time.strftime("%H:%M:%S", time.localtime(self.start_time))
        start_rss = self.start_memory.get("rss", 0)
        end_rss = self.end_memory.get("rss", 0)

        buf = [
            f"Name: {self.name}",
            f"PID: {self.pid}",
            f"Start: {start_ts}",
            f"Duration: {self.duration:.3f}s",
            f"Start RSS: {format_bytes(start_rss)}",
            f"End RSS: {format_bytes(end_rss)}",
            f"RSS delta: {format_bytes(self.rss_delta, show_sign=True)}",
            f"Avail memory: {format_bytes(current_available)}",
        ]
        return ' | '.join(buf)



# Contents in this file are referenced from the sphinx-generated docs.
# "magictoken" is used for markers as beginning and ending of example text.

import unittest
from numba.tests.support import captured_stdout


class DocsTypedSetUsageTest(unittest.TestCase):

    def test_ex_typed_set_from_cpython(self):
        with captured_stdout():
            # magictoken.ex_typed_set_from_cpython.begin
            import numpy as np
            from numba import njit
            from numba.core import types
            from numba.typed import Set

            # The Set.empty() constructs a typed set.
            # The key typed must be explicitly declared.
            s = Set.empty(types.unicode_type)

            # The typed-set can be used from the interpreter.
            s.add("One")
            s.add("Two")
            s.add("Three")

            # Here's a function that expects a typed-set as the argument
            @njit
            def remove_elem(s, elem):
                s.discard(elem)

            print(len(s))  # Out: 3

            # Call remove_elem(s) to remove an element in the typed-set.
            remove_elem(s, "Three")

            print(len(s))  # Out: 2
            # magictoken.ex_typed_set_from_cpython.end

        # Test
        np.testing.assert_equal(len(s), 2)


if __name__ == '__main__':
    unittest.main()



import functools
import operator
import warnings

import numpy as np

from numba import jit, typeof
from numba.core import cgutils, types, serialize, sigutils, errors
from numba.core.extending import (is_jitted, overload_attribute,
                                  overload_method, register_jitable,
                                  intrinsic)
from numba.core.typing import npydecl
from numba.core.typing.templates import AbstractTemplate, signature
from numba.cpython.unsafe.tuple import tuple_setitem
from numba.np.ufunc import _internal
from numba.np.ufunc.ufunc_base import UfuncBase, UfuncLowererBase
from numba.parfors import array_analysis
from numba.np.ufunc import ufuncbuilder
from numba.np import numpy_support
from typing import Callable
from llvmlite import ir
from numba.core.compiler_lock import global_compiler_lock


class UfuncAtIterator:

    def __init__(self, ufunc, a, a_ty, indices, indices_ty, b=None, b_ty=None):
        self.ufunc = ufunc
        self.a = a
        self.a_ty = a_ty
        self.indices = indices
        self.indices_ty = indices_ty
        self.b = b
        self.b_ty = b_ty

    def run(self, context, builder):
        self._prepare(context, builder)
        loop_indices = self.indexer.begin_loops()
        self._call_ufunc(context, builder, loop_indices)
        self.indexer.end_loops()

    def need_advanced_indexing(self):
        return isinstance(self.indices_ty, types.BaseTuple)

    def _prepare(self, context, builder):
        from numba.np.arrayobj import (normalize_indices, FancyIndexer,
                                       make_array, get_subspace_shape)

        a, indices = self.a, self.indices
        a_ty, indices_ty = self.a_ty, self.indices_ty

        zero = context.get_value_type(types.intp)(0)

        if self.b is not None:
            self.b_indice = cgutils.alloca_once_value(builder, zero)

        if self.need_advanced_indexing():
            indices = cgutils.unpack_tuple(builder, indices,
                                           count=len(indices_ty))
            index_types = indices_ty.types
            index_types, indices = normalize_indices(context, builder,
                                                     index_types, indices)
        else:
            indices = (indices,)
            index_types = (indices_ty,)
            index_types, indices = normalize_indices(context, builder,
                                                     index_types, indices)

        array_indices = []
        for i, idxty in enumerate(index_types):
            idx = indices[i]
            if isinstance(idxty, types.Array):
                idx_make = make_array(idxty)(context, builder, idx)
                array_indices.append((i, idx, idxty, idx_make))

        if array_indices:
            subspace_shape_tuple = \
                get_subspace_shape(context, builder, array_indices)
        else:
            subspace_shape_tuple = ()
        self.indexer = FancyIndexer(context, builder, a_ty, a,
                                    index_types, indices, subspace_shape_tuple)
        self.indexer.prepare()
        self.cres = self._compile_ufunc(context, builder)

    def _load_val(self, context, builder, loop_indices, array, array_ty):
        from numba.np.arrayobj import load_item
        shapes = cgutils.unpack_tuple(builder, array.shape)
        strides = cgutils.unpack_tuple(builder, array.strides)
        data = array.data

        ptr = cgutils.get_item_pointer2(context, builder, data, shapes, strides,
                                        array_ty.layout, loop_indices)
        val = load_item(context, builder, array_ty, ptr)
        return ptr, val

    def _load_flat(self, context, builder, indices, array, array_ty):
        idx = builder.load(indices)
        sig = array_ty.dtype(array_ty, types.intp)
        impl = context.get_function(operator.getitem, sig)
        val = impl(builder, (array, idx))

        # increment indices
        one = context.get_value_type(types.intp)(1)
        idx = builder.add(idx, one)
        builder.store(idx, indices)

        return None, val

    def _store_val(self, context, builder, array, array_ty, ptr, val):
        from numba.np.arrayobj import store_item
        fromty = self.cres.signature.return_type
        toty = array_ty.dtype
        val = context.cast(builder, val, fromty, toty)
        store_item(context, builder, array_ty, val, ptr)

    def _compile_ufunc(self, context, builder):
        ufunc = self.ufunc.key[0]

        if self.b is None:
            sig = (self.a_ty.dtype,)
        else:
            sig = (self.a_ty.dtype, self.b_ty.dtype)

        cres = ufunc.add(sig)
        context.add_linking_libs((cres.library,))
        return cres

    def _call_ufunc(self, context, builder, loop_indices):
        cres = self.cres
        a, a_ty = self.a, self.a_ty

        ptr, val = self._load_val(context, builder, loop_indices, a, a_ty)

        if self.b is None:
            args = (val,)
        else:
            b, b_ty, b_idx = self.b, self.b_ty, self.b_indice
            _, val_b = self._load_flat(context, builder, b_idx, b, b_ty)
            args = (val, val_b)

        res = context.call_internal(builder, cres.fndesc, cres.signature,
                                    args)
        self._store_val(context, builder, a, a_ty, ptr, res)


def make_dufunc_kernel(_dufunc):
    from numba.np import npyimpl

    class DUFuncKernel(npyimpl._Kernel):
        """
        npyimpl._Kernel subclass responsible for lowering a DUFunc kernel
        (element-wise function) inside a broadcast loop (which is
        generated by npyimpl.numpy_ufunc_kernel()).
        """
        dufunc = _dufunc

        def __init__(self, context, builder, outer_sig):
            super().__init__(context, builder, outer_sig)
            self.inner_sig, self.cres = self.dufunc.find_ewise_function(
                outer_sig.args)

    DUFuncKernel.__name__ += _dufunc.ufunc.__name__
    return DUFuncKernel


class DUFuncLowerer(UfuncLowererBase):
    '''Callable class responsible for lowering calls to a specific DUFunc.
    '''
    def __init__(self, dufunc):
        from numba.np import npyimpl
        super().__init__(dufunc,
                         make_dufunc_kernel,
                         npyimpl.numpy_ufunc_kernel)


class DUFunc(serialize.ReduceMixin, _internal._DUFunc, UfuncBase):
    """
    Dynamic universal function (DUFunc) intended to act like a normal
    Numpy ufunc, but capable of call-time (just-in-time) compilation
    of fast loops specialized to inputs.
    """
    # NOTE: __base_kwargs must be kept in synch with the kwlist in
    # _internal.c:dufunc_init()
    __base_kwargs = set(('identity', '_keepalive', 'nin', 'nout'))

    def __init__(self, py_func, identity=None, cache=False, targetoptions=None):
        if targetoptions is None:
            targetoptions = {}
        if is_jitted(py_func):
            py_func = py_func.py_func
        with ufuncbuilder._suppress_deprecation_warning_nopython_not_supplied():
            dispatcher = jit(_target='npyufunc',
                             cache=cache,
                             **targetoptions)(py_func)
        self._initialize(dispatcher, identity)
        functools.update_wrapper(self, py_func)

    def _initialize(self, dispatcher, identity):
        identity = ufuncbuilder.parse_identity(identity)
        super(DUFunc, self).__init__(dispatcher, identity=identity)
        # Loop over a copy of the keys instead of the keys themselves,
        # since we're changing the dictionary while looping.
        self.reorderable = (identity != _internal.PyUFunc_None)
        self.__name__ = dispatcher.py_func.__name__
        self.__doc__ = dispatcher.py_func.__doc__
        self._lower_me = DUFuncLowerer(self)
        self._install_cg()
        self._install_type()

    def _reduce_states(self):
        """
        NOTE: part of ReduceMixin protocol
        """
        siglist = list(self._dispatcher.overloads.keys())
        return dict(
            dispatcher=self._dispatcher,
            identity=self.identity,
            frozen=self._frozen,
            siglist=siglist,
        )

    @classmethod
    def _rebuild(cls, dispatcher, identity, frozen, siglist):
        """
        NOTE: part of ReduceMixin protocol
        """
        self = _internal._DUFunc.__new__(cls)
        self._initialize(dispatcher, identity)
        # Re-add signatures
        for sig in siglist:
            self.add(sig)
        if frozen:
            self.disable_compile()
        return self

    def build_ufunc(self):
        """
        For compatibility with the various *UFuncBuilder classes.
        """
        return self

    @property
    def targetoptions(self):
        return self._dispatcher.targetoptions

    @property
    def nin(self):
        return self.ufunc.nin

    @property
    def nout(self):
        return self.ufunc.nout

    @property
    def nargs(self):
        return self.ufunc.nargs

    @property
    def ntypes(self):
        return self.ufunc.ntypes

    @property
    def types(self):
        return self.ufunc.types

    @property
    def identity(self):
        return self.ufunc.identity

    @property
    def signature(self):
        return self.ufunc.signature

    def disable_compile(self):
        """
        Disable the compilation of new signatures at call time.
        """
        # If disabling compilation then there must be at least one signature
        assert len(self._dispatcher.overloads) > 0
        self._frozen = True

    def add(self, sig):
        """
        Compile the DUFunc for the given signature.
        """
        args, return_type = sigutils.normalize_signature(sig)
        return self._compile_for_argtys(args, return_type)

    def __call__(self, *args, **kws):
        """
        Allow any argument that has overridden __array_ufunc__ (NEP-18)
        to take control of DUFunc.__call__.
        """
        default = numpy_support.np.ndarray.__array_ufunc__

        for arg in args + tuple(kws.values()):
            if getattr(type(arg), "__array_ufunc__", default) is not default:
                output = arg.__array_ufunc__(self, "__call__", *args, **kws)
                if output is not NotImplemented:
                    return output
        else:
            return super().__call__(*args, **kws)

    def _compile_for_args(self, *args, **kws):
        nin = self.ufunc.nin
        if kws:
            if 'out' in kws:
                out = kws.pop('out')
                args += (out,)
            if kws:
                raise TypeError("unexpected keyword arguments to ufunc: %s"
                                % ", ".join(repr(k) for k in sorted(kws)))

        args_len = len(args)
        assert (args_len == nin) or (args_len == nin + self.ufunc.nout)
        assert not kws
        argtys = []
        for arg in args[:nin]:
            argty = typeof(arg)
            if isinstance(argty, types.Array):
                argty = argty.dtype
            else:
                # To avoid a mismatch in how Numba types scalar values as
                # opposed to Numpy, we need special logic for scalars.
                # For example, on 64-bit systems, numba.typeof(3) => int32, but
                # np.array(3).dtype => int64.

                # Note: this will not handle numpy "duckarrays" correctly,
                # including but not limited to those implementing `__array__`
                # and `__array_ufunc__`.
                argty = numpy_support.map_arrayscalar_type(arg)
            argtys.append(argty)
        return self._compile_for_argtys(tuple(argtys))

    @global_compiler_lock
    def _compile_for_argtys(self, argtys, return_type=None):
        """
        Given a tuple of argument types (these should be the array
        dtypes, and not the array types themselves), compile the
        element-wise function for those inputs, generate a UFunc loop
        wrapper, and register the loop with the Numpy ufunc object for
        this DUFunc.
        """
        if self._frozen:
            raise RuntimeError("compilation disabled for %s" % (self,))
        assert isinstance(argtys, tuple)
        if return_type is None:
            sig = argtys
        else:
            sig = return_type(*argtys)

        for k, cres in self._dispatcher.overloads.items():
            if argtys == k.args:
                msg = ("Compilation requested for previously compiled argument"
                       f" types ({argtys}). This has no effect and perhaps "
                       "indicates a bug in the calling code (compiling a "
                       "ufunc more than once for the same signature")
                warnings.warn(msg, errors.NumbaWarning)
                return cres

        cres, argtys, return_type = ufuncbuilder._compile_element_wise_function(
            self._dispatcher, self.targetoptions, sig)
        actual_sig = ufuncbuilder._finalize_ufunc_signature(
            cres, argtys, return_type)
        dtypenums, ptr, env = ufuncbuilder._build_element_wise_ufunc_wrapper(
            cres, actual_sig)
        self._add_loop(int(ptr), dtypenums)
        self._keepalive.append((ptr, cres.library, env))
        self._lower_me.libs.append(cres.library)
        return cres

    def match_signature(self, ewise_types, sig):
        return sig.args == ewise_types

    def _install_ufunc_attributes(self, template) -> None:

        def get_attr_fn(attr: str) -> Callable:

            def impl(ufunc):
                val = getattr(ufunc.key[0], attr)
                return lambda ufunc: val
            return impl

        # ntypes/types needs "at" to be a BoundFunction rather than a Function
        # But this fails as it cannot a weak reference to an ufunc due to NumPy
        # not setting the "tp_weaklistoffset" field. See:
        # https://github.com/numpy/numpy/blob/7fc72776b972bfbfdb909e4b15feb0308cf8adba/numpy/core/src/umath/ufunc_object.c#L6968-L6983  # noqa: E501

        at = types.Function(template)
        attributes = ('nin', 'nout', 'nargs', # 'ntypes', # 'types',
                      'identity', 'signature')
        for attr in attributes:
            attr_fn = get_attr_fn(attr)
            overload_attribute(at, attr)(attr_fn)

    def _install_ufunc_methods(self, template) -> None:
        self._install_ufunc_reduce(template)
        self._install_ufunc_reduceat(template)
        self._install_ufunc_at(template)

    def _install_ufunc_at(self, template) -> None:
        at = types.Function(template)

        @overload_method(at, 'at')
        def ol_at(ufunc, a, indices, b=None):
            warnings.warn("ufunc.at feature is experimental",
                          category=errors.NumbaExperimentalFeatureWarning)

            if not isinstance(a, types.Array):
                msg = 'The first argument "a" must be array-like'
                raise errors.NumbaTypeError(msg)

            indices_arr = isinstance(indices, types.Array)
            indices_list = isinstance(indices, types.List)
            indices_tuple = isinstance(indices, types.Tuple)
            indices_slice = isinstance(indices, types.SliceType)
            indices_scalar = not (indices_arr or indices_slice or indices_tuple)
            indices_empty_tuple = indices_tuple and len(indices) == 0
            b_array = isinstance(b, (types.Array, types.Sequence, types.List,
                                     types.Tuple))
            b_none = cgutils.is_nonelike(b)
            b_scalar = not (b_array or b_none)
            need_cast = any([indices_list])

            nin = self.ufunc.nin

            # missing second argument?
            if nin == 2 and cgutils.is_nonelike(b):
                raise errors.TypingError('second operand needed for ufunc')

            # extra second argument
            if nin == 1 and not cgutils.is_nonelike(b):
                msg = 'second operand provided when ufunc is unary'
                raise errors.TypingError(msg)

            if cgutils.is_nonelike(b):
                self.add((a.dtype,))
            elif b_scalar:
                self.add((a.dtype, b))
            else:
                self.add((a.dtype, b.dtype))

            def apply_ufunc_codegen(context, builder, sig, args):
                from numba.np.arrayobj import make_array

                if len(args) == 4:
                    _, aty, idxty, bty = sig.args
                    _, a, indices, b = args
                else:
                    _, aty, idxty, bty = sig.args + (None,)
                    _, a, indices, b = args + (None,)

                a = make_array(aty)(context, builder, a)
                at_iter = UfuncAtIterator(ufunc, a, aty, indices, idxty, b, bty)
                at_iter.run(context, builder)

            @intrinsic
            def apply_a_b_ufunc(typingctx, ufunc, a, indices, b):
                sig = types.none(ufunc, a, indices, b)
                return sig, apply_ufunc_codegen

            @intrinsic
            def apply_a_ufunc(typingctx, ufunc, a, indices):
                sig = types.none(ufunc, a, indices)
                return sig, apply_ufunc_codegen

            def impl_cast(ufunc, a, indices, b=None):
                if b_none:
                    return ufunc.at(a, np.asarray(indices))
                else:
                    return ufunc.at(a,
                                    np.asarray(indices),
                                    np.asarray(b))

            def impl_generic(ufunc, a, indices, b=None):
                if b_none:
                    apply_a_ufunc(ufunc, a, indices,)
                else:
                    b_ = np.asarray(b)
                    a_ = a[indices]
                    b_ = np.broadcast_to(b_, a_.shape)
                    apply_a_b_ufunc(ufunc, a, indices, b_.flat)

            def impl_indices_empty_b_scalar(ufunc, a, indices, b=None):
                a[()] = ufunc(a[()], b)

            def impl_scalar_scalar(ufunc, a, indices, b=None):
                if b_none:
                    a[indices] = ufunc(a[indices])
                else:
                    a[indices] = ufunc(a[indices], b)

            if need_cast:
                return impl_cast
            elif indices_empty_tuple and b_scalar:
                return impl_indices_empty_b_scalar
            elif indices_scalar and b_scalar:
                return impl_scalar_scalar
            else:
                return impl_generic

    def _install_ufunc_reduce(self, template) -> None:
        at = types.Function(template)

        @overload_method(at, 'reduce')
        def ol_reduce(ufunc, array, axis=0, dtype=None, initial=None):

            warnings.warn("ufunc.reduce feature is experimental",
                          category=errors.NumbaExperimentalFeatureWarning)

            if not isinstance(array, types.Array):
                msg = 'The first argument "array" must be array-like'
                raise errors.NumbaTypeError(msg)

            axis_int_tuple = isinstance(axis, types.UniTuple) and \
                isinstance(axis.dtype, types.Integer)
            axis_empty_tuple = isinstance(axis, types.Tuple) and len(axis) == 0
            axis_none = cgutils.is_nonelike(axis)

            identity_none = self.ufunc.identity is None
            ufunc_name = self.ufunc.__name__

            # In NumPy, a ufunc is reorderable if its identity type is **not**
            # PyUfunc_None.
            if not self.reorderable and axis_int_tuple and len(axis) > 1:
                msg = (f"reduction operation '{ufunc_name}' is not "
                       "reorderable, so at most one axis may be specified")
                raise errors.NumbaTypeError(msg)

            tup_init = (0,) * (array.ndim)
            tup_init_m1 = (0,) * (array.ndim - 1)
            nb_dtype = array.dtype if cgutils.is_nonelike(dtype) else dtype
            identity = self.identity

            id_none = cgutils.is_nonelike(identity)
            init_none = cgutils.is_nonelike(initial)

            @register_jitable
            def tuple_slice(tup, pos):
                # Same as
                # tup = tup[0 : pos] + tup[pos + 1:]
                s = tup_init_m1
                i = 0
                for j, e in enumerate(tup):
                    if j == pos:
                        continue
                    s = tuple_setitem(s, i, e)
                    i += 1
                return s

            @register_jitable
            def tuple_slice_append(tup, pos, val):
                # Same as
                # tup = tup[0 : pos] + val + tup[pos + 1:]
                s = tup_init
                i, j, sz = 0, 0, len(s)
                while j < sz:
                    if j == pos:
                        s = tuple_setitem(s, j, val)
                    else:
                        e = tup[i]
                        s = tuple_setitem(s, j, e)
                        i += 1
                    j += 1
                return s

            @intrinsic
            def compute_flat_idx(typingctx, strides, itemsize, idx, axis):
                sig = types.intp(strides, itemsize, idx, axis)
                len_idx = len(idx)

                def gen_block(builder, block_pos, block_name, bb_end, args):
                    strides, _, idx, _ = args
                    bb = builder.append_basic_block(name=block_name)

                    with builder.goto_block(bb):
                        zero = ir.IntType(64)(0)
                        flat_idx = zero

                        if block_pos == 0:
                            for i in range(1, len_idx):
                                stride = builder.extract_value(strides, i - 1)
                                idx_i = builder.extract_value(idx, i)
                                m = builder.mul(stride, idx_i)
                                flat_idx = builder.add(flat_idx, m)
                        elif 0 < block_pos < len_idx - 1:
                            for i in range(0, block_pos):
                                stride = builder.extract_value(strides, i)
                                idx_i = builder.extract_value(idx, i)
                                m = builder.mul(stride, idx_i)
                                flat_idx = builder.add(flat_idx, m)

                            for i in range(block_pos + 1, len_idx):
                                stride = builder.extract_value(strides, i - 1)
                                idx_i = builder.extract_value(idx, i)
                                m = builder.mul(stride, idx_i)
                                flat_idx = builder.add(flat_idx, m)
                        else:
                            for i in range(0, len_idx - 1):
                                stride = builder.extract_value(strides, i)
                                idx_i = builder.extract_value(idx, i)
                                m = builder.mul(stride, idx_i)
                                flat_idx = builder.add(flat_idx, m)

                        builder.branch(bb_end)

                    return bb, flat_idx

                def codegen(context, builder, sig, args):
                    strides, itemsize, idx, axis = args

                    bb = builder.basic_block
                    switch_end = builder.append_basic_block(name='axis_end')
                    l = []
                    for i in range(len_idx):
                        block, flat_idx = gen_block(builder, i, f"axis_{i}",
                                                    switch_end, args)
                        l.append((block, flat_idx))

                    with builder.goto_block(bb):
                        switch = builder.switch(axis, l[-1][0])
                        for i in range(len_idx):
                            switch.add_case(i, l[i][0])

                    builder.position_at_end(switch_end)
                    phi = builder.phi(l[0][1].type)
                    for block, value in l:
                        phi.add_incoming(value, block)
                    return builder.sdiv(phi, itemsize)

                return sig, codegen

            @register_jitable
            def fixup_axis(axis, ndim):
                ax = axis
                for i in range(len(axis)):
                    val = axis[i] + ndim if axis[i] < 0 else axis[i]
                    ax = tuple_setitem(ax, i, val)
                return ax

            @register_jitable
            def find_min(tup):
                idx, e = 0, tup[0]
                for i in range(len(tup)):
                    if tup[i] < e:
                        idx, e = i, tup[i]
                return idx, e

            def impl_1d(ufunc, array, axis=0, dtype=None, initial=None):
                if identity_none and initial is None and len(array) == 0:
                    msg = ('zero-size array to reduction operation '
                           f'{ufunc_name} which has no identity')
                    raise ValueError(msg)

                start = 0
                if init_none and id_none:
                    start = 1
                    r = array[0]
                elif init_none:
                    r = identity
                else:
                    r = initial

                sz = array.shape[0]
                for i in range(start, sz):
                    r = ufunc(r, array[i])
                return r

            def impl_nd_axis_int(ufunc,
                                 array,
                                 axis=0,
                                 dtype=None,
                                 initial=None):
                if axis is None:
                    raise ValueError("'axis' must be specified")

                if axis < 0:
                    axis += array.ndim

                if axis < 0 or axis >= array.ndim:
                    raise ValueError("Invalid axis")

                if identity_none and initial is None and array.shape[axis] == 0:
                    msg = ('zero-size array to reduction operation '
                           f'{ufunc_name} which has no identity')
                    raise ValueError(msg)

                # create result array
                shape = tuple_slice(array.shape, axis)

                if initial is None and identity is None:
                    r = np.empty(shape, dtype=nb_dtype)
                    for idx, _ in np.ndenumerate(r):
                        # shape[0:axis] + 0 + shape[axis:]
                        result_idx = tuple_slice_append(idx, axis, 0)
                        r[idx] = array[result_idx]
                elif initial is None and identity is not None:
                    # Checking if identity is not none is redundant but required
                    # compile this block
                    r = np.full(shape, fill_value=identity, dtype=nb_dtype)
                else:
                    r = np.full(shape, fill_value=initial, dtype=nb_dtype)

                # One approach to implement reduce is to remove the axis index
                # from the indexing tuple returned by "np.ndenumerate". For
                # instance, if idx = (X, Y, Z) and axis=1, the result index
                # is (X, Y).
                # Another way is to compute the result index using strides,
                # which is faster than manipulating tuples.
                view = r.ravel()
                if initial is None and identity is None:
                    for idx, val in np.ndenumerate(array):
                        if idx[axis] == 0:
                            continue
                        else:
                            flat_pos = compute_flat_idx(r.strides, r.itemsize,
                                                        idx, axis)
                            lhs, rhs = view[flat_pos], val
                            view[flat_pos] = ufunc(lhs, rhs)
                else:
                    for idx, val in np.ndenumerate(array):
                        if initial is None and identity is None and \
                                idx[axis] == 0:
                            continue
                        flat_pos = compute_flat_idx(r.strides, r.itemsize,
                                                    idx, axis)
                        lhs, rhs = view[flat_pos], val
                        view[flat_pos] = ufunc(lhs, rhs)
                return r

            def impl_nd_axis_tuple(ufunc,
                                   array,
                                   axis=0,
                                   dtype=None,
                                   initial=None):
                axis_ = fixup_axis(axis, array.ndim)
                for i in range(0, len(axis_)):
                    if axis_[i] < 0 or axis_[i] >= array.ndim:
                        raise ValueError("Invalid axis")

                    for j in range(i + 1, len(axis_)):
                        if axis_[i] == axis_[j]:
                            raise ValueError("duplicate value in 'axis'")

                min_idx, min_elem = find_min(axis_)
                r = ufunc.reduce(array,
                                 axis=min_elem,
                                 dtype=dtype,
                                 initial=initial)
                if len(axis) == 1:
                    return r
                elif len(axis) == 2:
                    return ufunc.reduce(r, axis=axis_[(min_idx + 1) % 2] - 1)
                else:
                    ax = axis_tup
                    for i in range(len(ax)):
                        if i != min_idx:
                            ax = tuple_setitem(ax, i, axis_[i])
                    return ufunc.reduce(r, axis=ax)

            def impl_axis_empty_tuple(ufunc,
                                      array,
                                      axis=0,
                                      dtype=None,
                                      initial=None):
                return array

            def impl_axis_none(ufunc,
                               array,
                               axis=0,
                               dtype=None,
                               initial=None):
                return ufunc.reduce(array, axis_tup, dtype, initial)

            if array.ndim == 1 and not axis_empty_tuple:
                return impl_1d
            elif axis_empty_tuple:
                # ufunc(array, axis=())
                return impl_axis_empty_tuple
            elif axis_none:
                # ufunc(array, axis=None)
                axis_tup = tuple(range(array.ndim))
                return impl_axis_none
            elif axis_int_tuple:
                # axis is tuple of integers
                # ufunc(array, axis=(1, 2, ...))
                axis_tup = (0,) * (len(axis) - 1)
                return impl_nd_axis_tuple
            elif axis == 0 or isinstance(axis, (types.Integer,
                                                types.Omitted,
                                                types.IntegerLiteral)):
                # axis is default value (0) or an integer
                # ufunc(array, axis=0)
                return impl_nd_axis_int

    def _install_ufunc_reduceat(self, template) -> None:
        at = types.Function(template)

        @overload_method(at, 'reduceat')
        def ol_reduceat(ufunc, array, indices, axis=0, dtype=None, out=None):

            warnings.warn("ufunc.reduceat feature is experimental",
                          category=errors.NumbaExperimentalFeatureWarning)

            if self.ufunc.nin != 2:
                msg = 'reduceat only supported for binary functions'
                raise errors.NumbaTypeError(msg)

            if not numpy_support.type_can_asarray(array):
                msg = 'The first argument "array" must be array-like'
                raise errors.NumbaTypeError(msg)

            if not numpy_support.type_can_asarray(indices):
                msg = 'The second argument "indices" must be array-like'
                raise errors.NumbaTypeError(msg)

            if not isinstance(axis, (types.Integer, types.Omitted,
                                     types.IntegerLiteral, int)):
                msg = '"axis" must be an integer'
                raise errors.NumbaTypeError(msg)

            out_none = cgutils.is_nonelike(out)
            if not out_none and not isinstance(out, types.Array):
                raise errors.NumbaTypeError('output must be an array')

            array_arr = isinstance(array, types.Array)
            array_ndim = array.ndim if array_arr else 0
            tup_m1 = (0,) * (array_ndim - 1)
            indices_arr = isinstance(indices, types.Array)
            dt = array.dtype if cgutils.is_nonelike(dtype) else dtype

            need_cast = not (array_arr and indices_arr)

            if indices_arr and indices.ndim != 1:
                msg = ('Expect "indices" array to have at most 1 dimension. '
                       f'Got {indices.ndim}')
                raise errors.NumbaValueError(msg)

            if need_cast:
                def impl_cast(ufunc, array, indices, axis=0, dtype=None, out=None):  # noqa: E501
                    return ufunc.reduceat(np.asarray(array),
                                          np.asarray(indices),
                                          axis,
                                          dtype=dtype,
                                          out=out)
                return impl_cast

            @register_jitable
            def tuple_slice(tup, pos):
                s = tup_m1
                i, j = 0, 0
                while i < len(tup):
                    if i != pos:
                        s = tuple_setitem(s, j, tup[i])
                        j += 1
                    i += 1
                return s

            # importing here as an import at the top scope brings unwanted
            # stuff. See numba/tests/test_import.py::test_no_impl_import
            from numba.np.arrayobj import generate_getitem_setitem_with_axis
            setitem = generate_getitem_setitem_with_axis(array.ndim, 'setitem')

            def impl(ufunc, array, indices, axis=0, dtype=None, out=None):
                sz = indices.shape[0]

                ax = axis
                if axis < 0:
                    axis += array_ndim

                if axis < 0 or axis >= array_ndim:
                    raise ValueError(f"axis {ax} is out of bounds for array "
                                     f"of dimension {array_ndim}")

                shape = tuple_setitem(array.shape, axis, sz)

                if not out_none and out.shape != shape:
                    # TODO: improve error message once #9524 gets merged
                    msg = ('operands could not be broadcast together with '
                           'remmaped shapes')
                    raise ValueError(msg)

                if out_none:
                    out = np.zeros(shape, dtype=dt)

                # short-circuit to avoid overflow on Windows
                if len(indices) == 0:
                    return out

                j = 0
                for i in range(len(indices) - 1):
                    if indices[i] < indices[i + 1]:
                        idx = np.arange(indices[i], indices[i + 1])
                        if array_ndim > 1:
                            arr_slice = np.take(array, idx, axis)
                        else:
                            arr_slice = array[idx]
                        arr_reduce = ufunc.reduce(arr_slice, axis=axis)
                        setitem(out, j, axis, arr_reduce)
                    elif indices[i] >= indices[i + 1]:
                        idx = indices[i]
                        arr_slice = np.take(array, idx, axis)
                        if array_ndim > 1:
                            _slice = tuple_slice(array.shape, axis)
                            arr_slice = arr_slice.reshape(_slice)
                        setitem(out, j, axis, arr_slice)
                    elif indices[i] >= sz or indices[i] < 0:
                        raise ValueError('Invalid value for indices')
                    j += 1

                # last index
                idx = np.arange(indices[-1], array.shape[axis])
                if array_ndim > 1:
                    arr_slice = np.take(array, idx, axis)
                else:
                    arr_slice = array[idx]
                arr_reduce = ufunc.reduce(arr_slice, axis)
                setitem(out, j, axis, arr_reduce)

                return out

            return impl

    def at(self, a, indices, b=None):
        # dynamic compile ufunc.at
        args = (a,) if cgutils.is_nonelike(b) else (a, b)
        argtys = (typeof(arg) for arg in args)
        ewise_types = tuple(arg.dtype if isinstance(arg, types.Array) else arg
                            for arg in argtys)

        if self.find_ewise_function(ewise_types) == (None, None):
            # cannot find a matching function and compilation is disabled
            if self._frozen:
                msg = "compilation disabled for %s.at(...)" % (self,)
                raise RuntimeError(msg)

            self._compile_for_args(*args)

        # all good, just dispatch to the function
        if cgutils.is_nonelike(b):
            return super().at(a, indices)
        else:
            return super().at(*(a, indices, b))

    def _install_type(self, typingctx=None):
        """Constructs and installs a typing class for a DUFunc object in the
        input typing context.  If no typing context is given, then
        _install_type() installs into the typing context of the
        dispatcher object (should be same default context used by
        jit() and njit()).
        """
        if typingctx is None:
            typingctx = self._dispatcher.targetdescr.typing_context
        _ty_cls = type('DUFuncTyping_' + self.ufunc.__name__,
                       (AbstractTemplate,),
                       dict(key=self, generic=self._type_me))
        typingctx.insert_user_function(self, _ty_cls)
        self._install_ufunc_attributes(_ty_cls)
        self._install_ufunc_methods(_ty_cls)

    def find_ewise_function(self, ewise_types):
        """
        Given a tuple of element-wise argument types, find a matching
        signature in the dispatcher.

        Return a 2-tuple containing the matching signature, and
        compilation result.  Will return two None's if no matching
        signature was found.
        """
        if self._frozen:
            # If we cannot compile, coerce to the best matching loop
            loop = numpy_support.ufunc_find_matching_loop(self, ewise_types)
            if loop is None:
                return None, None
            ewise_types = tuple(loop.inputs + loop.outputs)[:len(ewise_types)]
        for sig, cres in self._dispatcher.overloads.items():
            if sig.args == ewise_types:
                return sig, cres
        return None, None

    def _type_me(self, argtys, kwtys):
        """
        Implement AbstractTemplate.generic() for the typing class
        built by DUFunc._install_type().

        Return the call-site signature after either validating the
        element-wise signature or compiling for it.
        """
        assert not kwtys
        ufunc = self.ufunc
        _handle_inputs_result = npydecl.Numpy_rules_ufunc._handle_inputs(
            ufunc, argtys, kwtys)
        base_types, explicit_outputs, ndims, layout = _handle_inputs_result
        explicit_output_count = len(explicit_outputs)
        if explicit_output_count > 0:
            ewise_types = tuple(base_types[:-len(explicit_outputs)])
        else:
            ewise_types = tuple(base_types)
        sig, cres = self.find_ewise_function(ewise_types)
        if sig is None:
            # Matching element-wise signature was not found; must
            # compile.
            if self._frozen:
                raise errors.NumbaTypeError("cannot call %s with types %s"
                                            % (self, argtys))
            self._compile_for_argtys(ewise_types)
            sig, cres = self.find_ewise_function(ewise_types)
            assert sig is not None
        if explicit_output_count > 0:
            outtys = list(explicit_outputs)
        elif ufunc.nout == 1:
            if ndims > 0:
                outtys = [types.Array(sig.return_type, ndims, layout)]
            else:
                outtys = [sig.return_type]
        else:
            raise errors.NumbaNotImplementedError("typing gufuncs (nout > 1)")
        outtys.extend(argtys)
        return signature(*outtys)


array_analysis.MAP_TYPES.append(DUFunc)



# Although this module is not public API, `numba.core.Dispatcher` depends on
# `numba._dispatcher`, so it needs to be typed. The `compute_fingerprint` and
# `typeof_init` functions are annotated for the sake of completeness.

import array as _array
import mmap as _mmap

import numpy as np

from collections.abc import Callable
from typing import Any, type_check_only

from typing_extensions import Generic, ParamSpec, Protocol, TypeAlias, TypeVar

_SignatureT_co = TypeVar(
    "_SignatureT_co",
    bound=Callable[..., object],
    default=Callable[..., Any],
    covariant=True,
)
_ParamsT = ParamSpec("_ParamsT")
_ReturnT = TypeVar("_ReturnT")


@type_check_only
class _HasValue(Protocol):
    @property
    def value(self, /) -> object: ...


_Fingerprintable: TypeAlias = (
    None
    | bool
    | int
    | float
    | complex
    | bytes
    | bytearray
    | memoryview
    | _array.array
    | _mmap.mmap
    | _HasValue
    | np.generic
    | np.dtype
    | np.ndarray
    | tuple["_Fingerprintable", ...]
    | list["_Fingerprintable"]
    | set["_Fingerprintable"]
)

# NOTE: `Dispatcher` is not a generic type at runtime.
# undocumented
class Dispatcher(Generic[_SignatureT_co]):
    def __init__(
        self,
        tmaddrobj: int,
        argct: int,
        fold_args: bool,
        argnames: tuple[str, ...],
        defargs: tuple[object, ...],
        can_fallback: bool,
        has_stararg: bool = False,
        exact_match_required: bool = False,
        /,
    ) -> None: ...
    def __call__(
        self: Dispatcher[Callable[_ParamsT, _ReturnT]],
        /,
        *args: _ParamsT.args,
        **kwargs: _ParamsT.kwargs,
    ) -> _ReturnT: ...

# undocumented
def compute_fingerprint(val: _Fingerprintable, /) -> bytes: ...

# undocumented
def typeof_init(
    omittedarg_type: type[_HasValue],
    typecode_dict: dict[str, int],
    /,
) -> None: ...



"""
Testing C implementation of the numba sets
"""

import ctypes

from numba.tests.support import TestCase
from numba import _helperlib


ENTRY_PRESENT = 1
OK = 0
ERR_KEY_NOT_FOUND = -1
ERR_SET_MUTATED = -2
ERR_ITER_EXHAUSTED = -3
ERR_SET_EMPTY = -4
ERR_CMP_FAILED = -5


def to_bytes(key, key_size):
    if isinstance(key, str):
        key_bytes = (key_size - len(key)) * b'\0' + key.encode()
    else:
        key_bytes = key.to_bytes(key_size, 'big')
    return key_bytes


class SetIter(object):
    """An iterator for the `Set`.
    """
    def __init__(self, parent):
        self.parent = parent
        itsize = self.parent.tc.numba_set_iter_sizeof()
        self.it_state_buf = (ctypes.c_char_p * itsize)(0)
        self.it = ctypes.cast(self.it_state_buf, ctypes.c_void_p)
        self.parent.set_iter(self.it)

    def __iter__(self):
        return self

    def __next__(self):
        return self.parent.set_iter_next(self.it)


class Set(object):
    """A wrapper around the C-API to provide a minimal set object for
    testing.
    """
    def __init__(self, tc, key_size, allocated):
        """
        Parameters
        ----------
        tc : TestCase instance
        key_size : int
            byte size for the items
        allocated : int
            number of items to allocate for
        """
        self.tc = tc
        self.key_size = key_size
        self.setp = self.set_new(key_size, allocated)

    # The following methods implement part of the set API

    def __len__(self):
        return self.set_length()

    def __contains__(self, key):
        key_bytes = to_bytes(key, self.key_size)
        return self.set_contains(key_bytes=key_bytes)

    def __iter__(self):
        return SetIter(self)

    def __del__(self):
        self.tc.numba_set_free(self.setp)

    def add(self, key):
        key_bytes = to_bytes(key, self.key_size)
        self.set_add(key_bytes)

    def discard(self, key):
        key_bytes = to_bytes(key, self.key_size)
        self.set_discard(key_bytes)

    # The methods below are higher-level wrappers for the C-API wrappers

    def set_new(self, key_size, allocated):
        setp = ctypes.c_void_p()
        status = self.tc.numba_set_new(
            ctypes.byref(setp), key_size, allocated,
        )
        self.tc.assertEqual(status, OK)
        return setp

    def set_length(self):
        return self.tc.numba_set_length(self.setp)

    def set_add(self, key):
        status = self.tc.numba_set_add(self.setp, key, hash(key))
        self.tc.assertEqual(status, OK)

    def set_contains(self, key_bytes):
        hash_key = hash(key_bytes)
        ix = self.tc.numba_set_contains(
            self.setp, key_bytes, hash_key
        )
        return ix

    def set_discard(self, key_bytes):
        hash_key = hash(key_bytes)
        ix = self.tc.numba_set_discard(
            self.setp, key_bytes, hash_key
        )
        return ix

    def set_iter(self, itptr):
        self.tc.numba_set_iter(itptr, self.setp)

    def set_iter_next(self, itptr):
        bi = ctypes.c_void_p(0)
        status = self.tc.numba_set_iter_next(
            itptr, ctypes.byref(bi),
        )
        if status == ERR_SET_MUTATED:
            raise ValueError('set mutated')
        elif status == ERR_ITER_EXHAUSTED:
            raise StopIteration
        else:
            self.tc.assertGreaterEqual(status, 0)
            item = (ctypes.c_char * self.key_size).from_address(bi.value)
            return item.value


class TestSetImpl(TestCase):
    def setUp(self):
        """Bind to the c_helper library and provide the ctypes wrapper.
        """
        set_ty = ctypes.c_void_p
        hash_ty = ctypes.c_ssize_t
        setiter_ty = ctypes.c_void_p

        def wrap(name, restype, argtypes=()):
            proto = ctypes.CFUNCTYPE(restype, *argtypes)
            return proto(_helperlib.c_helpers[name])

        # numba_set_new(NB_set *setp, Py_ssize_t key_size, Py_ssize_t allocated)
        self.numba_set_new = wrap(
            'set_new',
            ctypes.c_int,
            [ctypes.POINTER(set_ty), ctypes.c_ssize_t, ctypes.c_ssize_t],
        )

        # numba_test_set()
        self.numba_test_set = wrap(
            'test_set',
            ctypes.c_int,
        )
        # numba_set_length(NB_set *setp)
        self.numba_set_length = wrap(
            'set_length',
            ctypes.c_int,
            [set_ty],
        )
        # numba_set_add(NB_set *setp, char *key, Py_ssize_t hash)
        self.numba_set_add = wrap(
            'set_add',
            ctypes.c_int,
            [set_ty, ctypes.c_char_p, hash_ty],
        )
        # numba_set_contains(NB_Set *setp, char *key, Py_ssize_t hash)
        self.numba_set_contains = wrap(
            'set_contains',
            ctypes.c_int,
            [set_ty, ctypes.c_char_p, hash_ty],
        )
        # numba_set_discard(NB_Set *setp, char *key, Py_ssize_t hash)
        self.numba_set_discard = wrap(
            'set_discard',
            ctypes.c_int,
            [set_ty, ctypes.c_char_p, hash_ty],
        )
        # numba_set_iter_sizeof()
        self.numba_set_iter_sizeof = wrap(
            'set_iter_sizeof',
            ctypes.c_size_t,
        )
        # numba_set_iter(NB_SetIter *it, NB_Set *setp)
        self.numba_set_iter = wrap(
            'set_iter',
            None,
            [
                setiter_ty,
                set_ty,
            ],
        )
        # numba_set_iter_next(NB_SetIter *it, const char **key_ptr)
        self.numba_set_iter_next = wrap(
            'set_iter_next',
            ctypes.c_int,
            [
                setiter_ty,                             # it
                ctypes.POINTER(ctypes.c_void_p),    # item_ptr
            ],
        )
        # numba_set_free(NB_Set *setp)
        self.numba_set_free = wrap(
            'set_free',
            None,
            [set_ty],
        )

    def test_simple_c_test(self):
        # Runs the basic test in C.
        ret = self.numba_test_set()
        self.assertEqual(ret, OK)

    def test_init_len(self):
        # Checks length at initialization
        s = Set(self, 8, 8)

        # length == zero
        self.assertEqual(len(s), 0)

    def test_add_check(self):
        # Test adding keys to a set

        for key in [3, 'a', "abc"]:
            s = Set(self, 8, 8)
            # Add key to the set
            s.add(key)
            # Length of set == 1
            self.assertEqual(len(s), 1)
            # Check key exists in set
            has_entry = s.__contains__(key)
            self.assertEqual(has_entry, True)

    def test_remove_check(self):
        # Test removal of keys from a set

        for key in [3, 'a', "abc"]:
            s = Set(self, 8, 8)
            # Add key to the set
            s.add(key)
            # Check key exists in set
            has_entry = s.__contains__(key)
            self.assertEqual(has_entry, True)
            # Discard the key
            s.discard(key)
            # Check key no longer exists in set
            has_entry = s.__contains__(key)
            self.assertEqual(has_entry, False)

    def test_set_iter(self):
        # Test iteration
        s = Set(self, 3, 8)
        nmax = 1000

        # Add elements to the set
        for i in range(100, nmax):
            s.add(f'{i}')

        keys = []
        # Iterate over the set
        for key in s:
            keys.append(key)

        # Check every key was present in the iteration
        self.assertEqual(len(s), len(keys))
        for i in range(100, nmax):
            self.assertIn(to_bytes(f'{i}', 3), keys)



"""
Compiler-side implementation of the sets.
"""
import ctypes
import operator

from enum import IntEnum

from llvmlite import ir
from numba import _helperlib

from numba.core.extending import (
    overload,
    overload_method,
    intrinsic,
    register_model,
    models,
    lower_builtin,
    make_attribute_wrapper,
)
from numba.core.imputils import iternext_impl, RefType
from numba.core import types, cgutils, typing
from numba.core.types import (
    SetType,
    SetIterableType,
    SetIteratorType,
    Type
)
from numba.core.errors import TypingError
from numba.typed.typedobjectutils import (_get_incref_decref,
                                          _container_get_data,
                                          _as_bytes, _cast,
                                          _get_equal)
from numba.core.imputils import impl_ret_borrowed


ll_set_type = cgutils.voidptr_t
ll_setiter_type = cgutils.voidptr_t
ll_voidptr_type = cgutils.voidptr_t
ll_status = cgutils.int32_t
ll_ssize_t = cgutils.intp_t
ll_hash = ll_ssize_t
ll_bytes = cgutils.voidptr_t


_meminfo_setptr = types.MemInfoPointer(types.voidptr)

# The following enums must match _setobject.c


class Status(IntEnum):
    """Status code for set operations.
    """
    ENTRY_PRESENT = 1
    OK = 0
    ERR_KEY_NOT_FOUND = -1
    ERR_SET_MUTATED = -2
    ERR_ITER_EXHAUSTED = -3
    ERR_SET_EMPTY = -4
    ERR_CMP_FAILED = -5


def new_set(key):
    """Construct a new set.

    Parameters
    ----------
    key : TypeRef
        Key type of the new set.
    """
    # With JIT disabled, ignore all arguments and return a Python set.
    return set()


@register_model(SetType)
class SetModel(models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ('meminfo', _meminfo_setptr),
            ('data', types.voidptr),   # ptr to the C set
        ]
        super(SetModel, self).__init__(dmm, fe_type, members)


@register_model(SetIterableType)
@register_model(SetIteratorType)
class SetIterModel(models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ('parent', fe_type.parent),  # reference to the set
            ('state', types.voidptr),    # iterator state in C code
        ]
        super(SetIterModel, self).__init__(dmm, fe_type, members)


# Make _parent available to make len simple
make_attribute_wrapper(SetIteratorType, "parent", "_parent")
make_attribute_wrapper(SetIterableType, "parent", "_parent")


def _raise_if_error(context, builder, status, msg):
    """Raise an internal error depending on the value of *status*
    """
    ok_status = status.type(int(Status.OK))
    with builder.if_then(builder.icmp_signed('!=', status, ok_status)):
        context.call_conv.return_user_exc(builder, RuntimeError, (msg,))


@intrinsic
def _as_meminfo(typingctx, setobj):
    """Returns the MemInfoPointer of a set.
    """
    if not isinstance(setobj, types.SetType):
        raise TypingError('expected *setobj* to be a SetType')

    def codegen(context, builder, sig, args):
        [ts] = sig.args
        [s] = args
        # Incref
        context.nrt.incref(builder, ts, s)
        ctor = cgutils.create_struct_proxy(ts)
        sstruct = ctor(context, builder, value=s)
        # Returns the plain MemInfo
        return sstruct.meminfo

    sig = _meminfo_setptr(setobj)
    return sig, codegen


@intrinsic
def _set_new_sized(typingctx, n_keys, keyty):
    """Wrap numba_set_new.

    Allocate a new set object.

    Parameters
    ----------
    keyty: Type
        Type of key, respectively.

    """
    resty = types.voidptr
    sig = resty(n_keys, keyty)

    def codegen(context, builder, sig, args):
        n_keys = builder.bitcast(args[0], ll_ssize_t)

        # Determine sizeof value type
        ll_val = context.get_data_type(keyty.instance_type)
        sz_val = context.get_abi_sizeof(ll_val)
        refsetp = cgutils.alloca_once(builder, ll_set_type, zfill=True)

        fnty = ir.FunctionType(
            ll_status,
            [ll_set_type.as_pointer(), ll_ssize_t, ll_ssize_t],
        )
        fn = cgutils.get_or_insert_function(builder.module, fnty,
                                            'numba_set_new_sized')

        status = builder.call(
            fn,
            [refsetp, ll_ssize_t(sz_val), n_keys],
        )
        _raise_if_error(
            context, builder, status,
            msg="Failed to allocate set",
        )
        setp = builder.load(refsetp)
        return setp

    return sig, codegen


def _call_set_free(context, builder, ptr):
    """Call numba_set_free(ptr)
    """
    fnty = ir.FunctionType(
        ir.VoidType(),
        [ll_set_type],
    )
    free = cgutils.get_or_insert_function(builder.module, fnty,
                                          'numba_set_free')
    builder.call(free, [ptr])


def _imp_dtor(context, module):
    """Define the dtor for set
    """
    llvoidptr = context.get_value_type(types.voidptr)
    llsize = context.get_value_type(types.uintp)
    fnty = ir.FunctionType(
        ir.VoidType(),
        [llvoidptr, llsize, llvoidptr],
    )
    fname = '_numba_set_dtor'
    fn = cgutils.get_or_insert_function(module, fnty, fname)

    if fn.is_declaration:
        # Set linkage
        fn.linkage = 'linkonce_odr'
        # Define
        builder = ir.IRBuilder(fn.append_basic_block())
        setp = builder.bitcast(fn.args[0], ll_set_type.as_pointer())
        s = builder.load(setp)
        _call_set_free(context, builder, s)
        builder.ret_void()

    return fn


@intrinsic
def _make_set(typingctx, keyty, ptr):
    """Make a set struct with the given *ptr*

    Parameters
    ----------
    keyty: Type
        Type of the key, respectively.
    ptr : llvm pointer value
        Points to the set object.
    """
    set_ty = types.SetType(keyty.instance_type)

    def codegen(context, builder, signature, args):
        [_, ptr] = args
        ctor = cgutils.create_struct_proxy(set_ty)
        sstruct = ctor(context, builder)
        sstruct.data = ptr

        alloc_size = context.get_abi_sizeof(
            context.get_value_type(types.voidptr),
        )
        dtor = _imp_dtor(context, builder.module)
        meminfo = context.nrt.meminfo_alloc_dtor(
            builder,
            context.get_constant(types.uintp, alloc_size),
            dtor,
        )

        data_pointer = context.nrt.meminfo_data(builder, meminfo)
        data_pointer = builder.bitcast(data_pointer, ll_set_type.as_pointer())
        builder.store(ptr, data_pointer)

        sstruct.meminfo = meminfo

        return sstruct._getvalue()

    sig = set_ty(keyty, ptr)
    return sig, codegen


@intrinsic
def _set_set_method_table(typingctx, setp, keyty):
    """Wrap numba_set_set_method_table
    """
    resty = types.void
    sig = resty(setp, keyty)

    def codegen(context, builder, sig, args):
        tablety = ir.LiteralStructType([
            ll_voidptr_type,  # equal
            ll_voidptr_type,  # key incref
            ll_voidptr_type,  # key decref
        ])
        setmethod_fnty = ir.FunctionType(
            ir.VoidType(),
            [ll_set_type, tablety.as_pointer()]
        )
        setmethod_fn = ir.Function(
            builder.module,
            setmethod_fnty,
            name='numba_set_set_method_table',
        )
        setp = args[0]
        table = cgutils.alloca_once(builder, tablety, zfill=True)

        key_equal_ptr = cgutils.gep_inbounds(builder, table, 0, 0)
        key_incref_ptr = cgutils.gep_inbounds(builder, table, 0, 1)
        key_decref_ptr = cgutils.gep_inbounds(builder, table, 0, 2)

        dm_key = context.data_model_manager[keyty.instance_type]
        if dm_key.contains_nrt_meminfo():
            equal = _get_equal(context, builder.module, dm_key, 'set_key')
            key_incref, key_decref = _get_incref_decref(
                context, builder.module, dm_key, 'set_key'
            )
            builder.store(
                builder.bitcast(equal, key_equal_ptr.type.pointee),
                key_equal_ptr,
            )
            builder.store(
                builder.bitcast(key_incref, key_incref_ptr.type.pointee),
                key_incref_ptr,
            )
            builder.store(
                builder.bitcast(key_decref, key_decref_ptr.type.pointee),
                key_decref_ptr,
            )

        builder.call(setmethod_fn, [setp, table])

    return sig, codegen


@overload(new_set)
def impl_new_set(value, n_keys=8):
    """Creates a new set with *value* as the type
    of the set value.
    """
    if not isinstance(value, Type):
        raise TypeError("expecting *value* to be a numba Type")

    keyty = value

    def imp(value, n_keys=8):
        setp = _set_new_sized(n_keys, keyty)
        _set_set_method_table(setp, keyty)
        s = _make_set(keyty, setp)
        return s

    return imp


@intrinsic
def _set_length(typingctx, s):
    """Wrap numba_set_length

    Returns the length of the set.
    """
    resty = types.intp
    sig = resty(s)

    def codegen(context, builder, sig, args):
        fnty = ir.FunctionType(
            ll_ssize_t,
            [ll_set_type],
        )
        fn = cgutils.get_or_insert_function(builder.module, fnty,
                                            'numba_set_length')
        [s] = args
        [ts] = sig.args
        setp = _container_get_data(context, builder, ts, s)
        n = builder.call(fn, [setp])
        return n

    return sig, codegen


@overload(len)
def impl_len(setp):
    """len(set)
    """
    if not isinstance(setp, types.SetType):
        return

    def impl(setp):
        return _set_length(setp)

    return impl


@overload(len)
def impl_len_iters(s):
    if not isinstance(s, (SetIterableType)):
        return

    def impl(s):
        return _set_length(s._parent)

    return impl


@intrinsic
def _set_add(typingctx, setp, key, hash_key):
    """Wrap numba_set_add
    """
    resty = types.int32
    sig = resty(setp, setp.key_type, types.intp)

    def codegen(context, builder, sig, args):
        fnty = ir.FunctionType(
            ll_status,
            [ll_set_type, ll_bytes, ll_hash],
        )
        [s, key, hash_key] = args
        [set_ty, keyty, hash_key_ty] = sig.args
        fn = cgutils.get_or_insert_function(builder.module, fnty,
                                            'numba_set_add')

        key_datamodel = context.data_model_manager[keyty]

        key_data = key_datamodel.as_data(builder, key)
        ptr_key = cgutils.alloca_once_value(builder, key_data)
        cgutils.memset_padding(builder, ptr_key)

        set_ptr = _container_get_data(context, builder, set_ty, s)
        status = builder.call(
            fn,
            [
                set_ptr,
                _as_bytes(builder, ptr_key),
                hash_key
            ],
        )
        return status

    return sig, codegen


@intrinsic
def _set_contains(typingctx, s, key, hash_key):
    """Wrap numba_set_contains

    Returns intp
    """
    resty = types.intp
    sig = resty(s, key, hash_key)

    def codegen(context, builder, sig, args):
        fnty = ir.FunctionType(
            ll_ssize_t,
            [ll_set_type, ll_bytes, ll_hash],
        )
        [set_ty, keyty, hash_keyty] = sig.args
        [s, key, hash_key] = args
        fn = cgutils.get_or_insert_function(builder.module, fnty,
                                            'numba_set_contains')

        dm_key = context.data_model_manager[keyty]

        data_key = dm_key.as_data(builder, key)
        ptr_key = cgutils.alloca_once_value(builder, data_key)
        cgutils.memset_padding(builder, ptr_key)

        setp = _container_get_data(context, builder, set_ty, s)
        ix = builder.call(
            fn,
            [
                setp,
                _as_bytes(builder, ptr_key),
                hash_key
            ],
        )
        return ix

    return sig, codegen


@intrinsic
def _from_meminfo(typingctx, mi, settyperef):
    """Recreate a set from a MemInfoPointer
    """
    if mi != _meminfo_setptr:
        raise TypingError('expected a MemInfoPointer for set.')
    settype = settyperef.instance_type
    if not isinstance(settype, SetType):
        raise TypingError('expected a {}'.format(SetType))

    def codegen(context, builder, sig, args):
        [tmi, tdref] = sig.args
        setty = tdref.instance_type
        [mi, _] = args

        ctor = cgutils.create_struct_proxy(setty)
        sstruct = ctor(context, builder)

        data_pointer = context.nrt.meminfo_data(builder, mi)
        data_pointer = builder.bitcast(data_pointer, ll_set_type.as_pointer())

        sstruct.data = builder.load(data_pointer)
        sstruct.meminfo = mi

        return impl_ret_borrowed(
            context,
            builder,
            settype,
            sstruct._getvalue(),
        )

    sig = settype(mi, settyperef)
    return sig, codegen


@intrinsic
def _set_discard(typingctx, s, key, hash_key):
    """Wrap numba_set_discard
    """
    resty = types.int32
    sig = resty(s, s.key_type, types.intp)

    def codegen(context, builder, sig, args):
        fnty = ir.FunctionType(
            ll_status,
            [ll_set_type, ll_bytes, ll_hash],
        )
        [s, key, hash_key] = args
        [setp_ty, keyty, hash_key_ty] = sig.args
        fn = cgutils.get_or_insert_function(builder.module, fnty,
                                            'numba_set_discard')

        key_datamodel = context.data_model_manager[keyty]

        key_data = key_datamodel.as_data(builder, key)

        ptr_key = cgutils.alloca_once_value(builder, key_data)
        cgutils.memset_padding(builder, ptr_key)

        setp = _container_get_data(context, builder, setp_ty, s)
        status = builder.call(
            fn,
            [
                setp,
                _as_bytes(builder, ptr_key),
                hash_key
            ],
        )
        return status

    return sig, codegen


def _iterator_codegen(resty):
    """The common codegen for iterator intrinsics.

    Populates the iterator struct and increfs.
    """

    def codegen(context, builder, sig, args):
        [s] = args
        [set_ty] = sig.args
        iterhelper = context.make_helper(builder, resty)
        iterhelper.parent = s
        iterhelper.state = iterhelper.state.type(None)
        return impl_ret_borrowed(
            context,
            builder,
            resty,
            iterhelper._getvalue(),
        )

    return codegen


@overload_method(types.SetType, 'add')
def impl_set_add(s, key):
    if not isinstance(s, types.SetType):
        return

    keyty = s.key_type

    def impl(s, key):
        casted_key = _cast(key, keyty)
        status = _set_add(s, casted_key, hash(casted_key))
        if status == 0:
            return

    if s.is_precise():
        # Handle the precise case.
        return impl
    else:
        # Handle the imprecise case.
        s = s.refine(key)
        # Re-bind the key type to match the arguments.
        keyty = s.key_type
        # Create the signature that we wanted this impl to have.
        sig = typing.signature(types.void, s, keyty)
        return sig, impl


@overload_method(types.SetType, 'discard')
def impl_discard(s, key):
    if not isinstance(s, types.SetType):
        return

    keyty = s.key_type

    def impl(s, key):
        castedkey = _cast(key, keyty)
        return _set_discard(s, castedkey, hash(castedkey))

    return impl


@overload(operator.contains)
def impl_contains(s, key):
    if not isinstance(s, types.SetType):
        return

    keyty = s.key_type

    def impl(s, key):
        key = _cast(key, keyty)
        ix = _set_contains(s, key, hash(key))
        return ix
    return impl


@overload(operator.eq)
def impl_equal(set_a, set_b):
    if not isinstance(set_a, types.SetType):
        return
    if not isinstance(set_b, types.SetType):
        # If RHS is not a set, always returns False
        def impl_type_mismatch(set_a, set_b):
            return False
        return impl_type_mismatch

    otherkeyty = set_b.key_type

    def impl_type_matched(set_a, set_b):
        if len(set_a) != len(set_b):
            return False
        for element in set_a:
            kb = _cast(element, otherkeyty)
            ix = _set_contains(set_b, kb, hash(kb))
            if ix == 0:
                return False
        return True

    return impl_type_matched


@overload(operator.ne)
def impl_not_equal(da, db):
    if not isinstance(da, types.SetType):
        return

    def impl(da, db):
        return not (da == db)

    return impl


@overload_method(types.SetType, 'copy')
def impl_copy(s):
    if not isinstance(s, types.SetType):
        return

    key_type = s.key_type

    def impl(s):
        newd = new_set(key_type, n_keys=len(s))
        for k in s:
            newd.add(k)
        return newd

    return impl


@lower_builtin('getiter', types.SetIterableType)
def impl_iterable_getiter(context, builder, sig, args):
    iterablety = sig.args[0]
    it = context.make_helper(builder, iterablety.iterator_type, args[0])

    fnty = ir.FunctionType(
        ir.VoidType(),
        [ll_setiter_type, ll_set_type],
    )

    fn = cgutils.get_or_insert_function(builder.module, fnty,
                                        'numba_set_iter')

    proto = ctypes.CFUNCTYPE(ctypes.c_size_t)
    setiter_sizeof = proto(_helperlib.c_helpers['set_iter_sizeof'])
    state_type = ir.ArrayType(ir.IntType(8), setiter_sizeof())

    pstate = cgutils.alloca_once(builder, state_type, zfill=True)
    it.state = _as_bytes(builder, pstate)

    setp = _container_get_data(context, builder, iterablety.parent, it.parent)
    builder.call(fn, [it.state, setp])
    return impl_ret_borrowed(
        context,
        builder,
        sig.return_type,
        it._getvalue(),
    )


@lower_builtin('getiter', types.SetType)
def impl_set_getiter(context, builder, sig, args):
    [setty] = sig.args
    [s] = args
    iterablety = types.SetIterableType(setty)
    it = context.make_helper(builder, iterablety.iterator_type)

    fnty = ir.FunctionType(
        ir.VoidType(),
        [ll_setiter_type, ll_set_type],
    )

    fn = cgutils.get_or_insert_function(builder.module, fnty, 'numba_set_iter')

    proto = ctypes.CFUNCTYPE(ctypes.c_size_t)
    setiter_sizeof = proto(_helperlib.c_helpers['set_iter_sizeof'])
    state_type = ir.ArrayType(ir.IntType(8), setiter_sizeof())

    pstate = cgutils.alloca_once(builder, state_type, zfill=True)
    it.state = _as_bytes(builder, pstate)
    it.parent = s

    setp = _container_get_data(context, builder, iterablety.parent, args[0])
    builder.call(fn, [it.state, setp])
    return impl_ret_borrowed(
        context,
        builder,
        sig.return_type,
        it._getvalue(),
    )


@lower_builtin('iternext', types.SetIteratorType)
@iternext_impl(RefType.BORROWED)
def impl_iterator_iternext(context, builder, sig, args, result):
    iter_type = sig.args[0]
    it = context.make_helper(builder, iter_type, args[0])

    p2p_bytes = ll_bytes.as_pointer()

    iternext_fnty = ir.FunctionType(
        ll_status,
        [ll_bytes, p2p_bytes]
    )
    iternext = cgutils.get_or_insert_function(
        builder.module, iternext_fnty, 'numba_set_iter_next',
    )
    key_raw_ptr = cgutils.alloca_once(builder, ll_bytes)

    status = builder.call(iternext, (it.state, key_raw_ptr))

    is_valid = builder.icmp_unsigned('==', status, status.type(0))
    result.set_valid(is_valid)

    with builder.if_then(is_valid):
        keyty = iter_type.parent.key_type

        key_datamodel = context.data_model_manager[keyty]

        key_ptr = builder.bitcast(
            builder.load(key_raw_ptr),
            key_datamodel.get_data_type().as_pointer(),
        )
        key = key_datamodel.load_from_data_pointer(builder, key_ptr)

        # All set iterators use this common implementation.
        # Their differences are resolved here.
        if isinstance(iter_type.iterable, SetIterableType):
            result.yield_(key)
        else:
            # unreachable
            raise AssertionError('unknown type: {}'.format(iter_type.iterable))



"""
Implement Dominance-Fronter-based SSA by Choi et al described in Inria SSA book

References:

- Static Single Assignment Book by Inria
  http://ssabook.gforge.inria.fr/latest/book.pdf
- Choi et al. Incremental computation of static single assignment form.
"""
import logging
import operator
import warnings
from functools import reduce
from copy import copy
from collections import defaultdict

from numba import config
from numba.core import ir, ir_utils, errors
from numba.core.utils import OrderedSet, _lazy_pformat
from numba.core.analysis import compute_cfg_from_blocks


_logger = logging.getLogger(__name__)


def reconstruct_ssa(func_ir):
    """Apply SSA reconstruction algorithm on the given IR.

    Produces minimal SSA using Choi et al algorithm.
    """
    func_ir.blocks = _run_ssa(func_ir.blocks)

    return func_ir


class _CacheListVars:
    def __init__(self):
        self._saved = {}

    def get(self, inst):
        got = self._saved.get(inst)
        if got is None:
            self._saved[inst] = got = inst.list_vars()
        return got


def _run_ssa(blocks):
    """Run SSA reconstruction on IR blocks of a function.
    """
    if not blocks:
        # Empty blocks?
        return {}
    # Run CFG on the blocks
    cfg = compute_cfg_from_blocks(blocks)
    df_plus = _iterated_domfronts(cfg)
    # Find SSA violators
    violators = _find_defs_violators(blocks, cfg)
    # Make cache for .list_vars()
    cache_list_vars = _CacheListVars()

    # Process one SSA-violating variable at a time
    for varname in violators:
        _logger.debug(
            "Fix SSA violator on var %s", varname,
        )
        # Fix up the LHS
        # Put fresh variables for all assignments to the variable
        blocks, defmap = _fresh_vars(blocks, varname)
        _logger.debug("Replaced assignments: %s", _lazy_pformat(defmap))
        # Fix up the RHS
        # Re-associate the variable uses with the reaching definition
        blocks = _fix_ssa_vars(blocks, varname, defmap, cfg, df_plus,
                               cache_list_vars)

    # Post-condition checks.
    # CFG invariant
    cfg_post = compute_cfg_from_blocks(blocks)
    if cfg_post != cfg:
        raise errors.CompilerError("CFG mutated in SSA pass")
    return blocks


def _fix_ssa_vars(blocks, varname, defmap, cfg, df_plus, cache_list_vars):
    """Rewrite all uses to ``varname`` given the definition map
    """
    states = _make_states(blocks)
    states['varname'] = varname
    states['defmap'] = defmap
    states['phimap'] = phimap = defaultdict(list)
    states['cfg'] = cfg
    states['phi_locations'] = _compute_phi_locations(df_plus, defmap)
    newblocks = _run_block_rewrite(blocks, states, _FixSSAVars(cache_list_vars))
    # insert phi nodes
    for label, philist in phimap.items():
        curblk = newblocks[label]
        # Prepend PHI nodes to the block
        curblk.body = philist + curblk.body
    return newblocks


def _iterated_domfronts(cfg):
    """Compute the iterated dominance frontiers (DF+ in literatures).

    Returns a dictionary which maps block label to the set of labels of its
    iterated dominance frontiers.
    """
    domfronts = {k: set(vs) for k, vs in cfg.dominance_frontier().items()}
    keep_going = True
    while keep_going:
        keep_going = False
        for k, vs in domfronts.items():
            inner = reduce(operator.or_, [domfronts[v] for v in vs], set())
            if inner.difference(vs):
                vs |= inner
                keep_going = True
    return domfronts


def _compute_phi_locations(iterated_df, defmap):
    # See basic algorithm in Ch 4.1 in Inria SSA Book
    # Compute DF+(defs)
    # DF of all DFs is the union of all DFs
    phi_locations = set()
    for deflabel, defstmts in defmap.items():
        if defstmts:
            phi_locations |= iterated_df[deflabel]
    return phi_locations


def _fresh_vars(blocks, varname):
    """Rewrite to put fresh variable names
    """
    states = _make_states(blocks)
    states['varname'] = varname
    states['defmap'] = defmap = defaultdict(list)
    newblocks = _run_block_rewrite(blocks, states, _FreshVarHandler())
    return newblocks, defmap


def _get_scope(blocks):
    first, *_ = blocks.values()
    return first.scope


def _find_defs_violators(blocks, cfg):
    """
    Returns
    -------
    res : Set[str]
        The SSA violators in a dictionary of variable names.
    """
    defs = defaultdict(list)
    uses = defaultdict(set)
    states = dict(defs=defs, uses=uses)
    _run_block_analysis(blocks, states, _GatherDefsHandler())
    _logger.debug("defs %s", _lazy_pformat(defs))
    # Gather violators by number of definitions.
    # The violators are added by the order that they are seen and the algorithm
    # scan from the first to the last basic-block as they occur in bytecode.
    violators = OrderedSet([k for k, vs in defs.items() if len(vs) > 1])
    # Gather violators by uses not dominated by the one def
    doms = cfg.dominators()
    for k, use_blocks in uses.items():
        if k not in violators:
            for label in use_blocks:
                dom = doms[label]
                def_labels = {label for _assign, label in defs[k] }
                if not def_labels.intersection(dom):
                    violators.add(k)
                    break
    _logger.debug("SSA violators %s", _lazy_pformat(violators))
    return violators


def _run_block_analysis(blocks, states, handler):
    for label, blk in blocks.items():
        _logger.debug("==== SSA block analysis pass on %s", label)
        states['label'] = label
        for _ in _run_ssa_block_pass(states, blk, handler):
            pass


def _run_block_rewrite(blocks, states, handler):
    newblocks = {}
    for label, blk in blocks.items():
        _logger.debug("==== SSA block rewrite pass on %s", label)
        newblk = ir.Block(scope=blk.scope, loc=blk.loc)

        newbody = []
        states['label'] = label
        states['block'] = blk
        for stmt in _run_ssa_block_pass(states, blk, handler):
            assert stmt is not None
            newbody.append(stmt)
        newblk.body = newbody
        newblocks[label] = newblk
    return newblocks


def _make_states(blocks):
    return dict(
        scope=_get_scope(blocks),
    )


def _run_ssa_block_pass(states, blk, handler):
    _logger.debug("Running %s", handler)
    for stmt in blk.body:
        _logger.debug("on stmt: %s", stmt)
        if isinstance(stmt, ir.Assign):
            ret = handler.on_assign(states, stmt)
        else:
            ret = handler.on_other(states, stmt)
        if ret is not stmt and ret is not None:
            _logger.debug("replaced with: %s", ret)
        yield ret


class _BaseHandler:
    """A base handler for all the passes used here for the SSA algorithm.
    """
    def on_assign(self, states, assign):
        """
        Called when the pass sees an ``ir.Assign``.

        Subclasses should override this for custom behavior

        Parameters
        -----------
        states : dict
        assign : numba.ir.Assign

        Returns
        -------
        stmt : numba.ir.Assign or None
            For rewrite passes, the return value is used as the replacement
            for the given statement.
        """

    def on_other(self, states, stmt):
        """
        Called when the pass sees an ``ir.Stmt`` that's not an assignment.

        Subclasses should override this for custom behavior

        Parameters
        -----------
        states : dict
        assign : numba.ir.Stmt

        Returns
        -------
        stmt : numba.ir.Stmt or None
            For rewrite passes, the return value is used as the replacement
            for the given statement.
        """


class _GatherDefsHandler(_BaseHandler):
    """Find all defs and uses of variable in each block

    ``states["label"]`` is a int; label of the current block
    ``states["defs"]`` is a Mapping[str, List[Tuple[ir.Assign, int]]]:
        - a mapping of the name of the assignee variable to the assignment
          IR node and the block label.
    ``states["uses"]`` is a Mapping[Set[int]]
    """
    def on_assign(self, states, assign):
        # keep track of assignment and the block
        states["defs"][assign.target.name].append((assign, states["label"]))
        # keep track of uses
        for var in assign.list_vars():
            k = var.name
            if k != assign.target.name:
                states["uses"][k].add(states["label"])

    def on_other(self, states, stmt):
        # keep track of uses
        for var in stmt.list_vars():
            k = var.name
            states["uses"][k].add(states["label"])


class UndefinedVariable:
    def __init__(self):
        raise NotImplementedError("Not intended for instantiation")

    target = ir.UNDEFINED


class _FreshVarHandler(_BaseHandler):
    """Replaces assignment target with new fresh variables.
    """
    def on_assign(self, states, assign):
        if assign.target.name == states['varname']:
            scope = states['scope']
            defmap = states['defmap']
            # Allow first assignment to retain the name
            if len(defmap) == 0:
                newtarget = assign.target
                _logger.debug("first assign: %s", newtarget)
                if newtarget.name not in scope.localvars:
                    wmsg = f"variable {newtarget.name!r} is not in scope."
                    warnings.warn(errors.NumbaIRAssumptionWarning(wmsg,
                                  loc=assign.loc))
            else:
                newtarget = scope.redefine(assign.target.name, loc=assign.loc)
            assign = ir.Assign(
                target=newtarget,
                value=assign.value,
                loc=assign.loc
            )
            defmap[states['label']].append(assign)
        return assign

    def on_other(self, states, stmt):
        return stmt


class _FixSSAVars(_BaseHandler):
    """Replace variable uses in IR nodes to the correct reaching variable
    and introduce Phi nodes if necessary. This class contains the core of
    the SSA reconstruction algorithm.

    See Ch 5 of the Inria SSA book for reference. The method names used here
    are similar to the names used in the pseudocode in the book.
    """

    def __init__(self, cache_list_vars):
        self._cache_list_vars = cache_list_vars

    def on_assign(self, states, assign):
        rhs = assign.value
        if isinstance(rhs, ir.Inst):
            newdef = self._fix_var(
                states, assign, self._cache_list_vars.get(assign.value),
            )
            # Has a replacement that is not the current variable
            if newdef is not None and newdef.target is not ir.UNDEFINED:
                if states['varname'] != newdef.target.name:
                    replmap = {states['varname']: newdef.target}
                    rhs = copy(rhs)

                    ir_utils.replace_vars_inner(rhs, replmap)
                    return ir.Assign(
                        target=assign.target,
                        value=rhs,
                        loc=assign.loc,
                    )
        elif isinstance(rhs, ir.Var):
            newdef = self._fix_var(states, assign, [rhs])
            # Has a replacement that is not the current variable
            if newdef is not None and newdef.target is not ir.UNDEFINED:
                if states['varname'] != newdef.target.name:
                    return ir.Assign(
                        target=assign.target,
                        value=newdef.target,
                        loc=assign.loc,
                    )

        return assign

    def on_other(self, states, stmt):
        newdef = self._fix_var(
            states, stmt, self._cache_list_vars.get(stmt),
        )
        if newdef is not None and newdef.target is not ir.UNDEFINED:
            if states['varname'] != newdef.target.name:
                replmap = {states['varname']: newdef.target}
                stmt = copy(stmt)
                ir_utils.replace_vars_stmt(stmt, replmap)
        return stmt

    def _fix_var(self, states, stmt, used_vars):
        """Fix all variable uses in ``used_vars``.
        """
        varnames = [k.name for k in used_vars]
        phivar = states['varname']
        if phivar in varnames:
            return self._find_def(states, stmt)

    def _find_def(self, states, stmt):
        """Find definition of ``stmt`` for the statement ``stmt``
        """
        _logger.debug("find_def var=%r stmt=%s", states['varname'], stmt)
        selected_def = None
        label = states['label']
        local_defs = states['defmap'][label]
        local_phis = states['phimap'][label]
        block = states['block']

        cur_pos = self._stmt_index(stmt, block)
        for defstmt in reversed(local_defs):
            # Phi nodes have no index
            def_pos = self._stmt_index(defstmt, block, stop=cur_pos)
            if def_pos < cur_pos:
                selected_def = defstmt
                break
            # Maybe it's a PHI
            elif defstmt in local_phis:
                selected_def = local_phis[-1]
                break

        if selected_def is None:
            selected_def = self._find_def_from_top(
                states, label, loc=stmt.loc,
            )
        return selected_def

    def _find_def_from_top(self, states, label, loc):
        """Find definition reaching block of ``label``.

        This method would look at all dominance frontiers.
        Insert phi node if necessary.
        """
        _logger.debug("find_def_from_top label %r", label)
        cfg = states['cfg']
        defmap = states['defmap']
        phimap = states['phimap']
        phi_locations = states['phi_locations']

        if label in phi_locations:
            scope = states['scope']
            loc = states['block'].loc
            # fresh variable
            freshvar = scope.redefine(states['varname'], loc=loc)
            # insert phi
            phinode = ir.Assign(
                target=freshvar,
                value=ir.Expr.phi(loc=loc),
                loc=loc,
            )
            _logger.debug("insert phi node %s at %s", phinode, label)
            defmap[label].insert(0, phinode)
            phimap[label].append(phinode)
            # Find incoming values for the Phi node
            for pred, _ in cfg.predecessors(label):
                incoming_def = self._find_def_from_bottom(
                    states, pred, loc=loc,
                )
                _logger.debug("incoming_def %s", incoming_def)
                phinode.value.incoming_values.append(incoming_def.target)
                phinode.value.incoming_blocks.append(pred)
            return phinode
        else:
            idom = cfg.immediate_dominators()[label]
            if idom == label:
                # We have searched to the top of the idom tree.
                # Since we still cannot find a definition,
                # we will warn.
                _warn_about_uninitialized_variable(states['varname'], loc)
                return UndefinedVariable
            _logger.debug("idom %s from label %s", idom, label)
            return self._find_def_from_bottom(states, idom, loc=loc)

    def _find_def_from_bottom(self, states, label, loc):
        """Find definition from within the block at ``label``.
        """
        _logger.debug("find_def_from_bottom label %r", label)
        defmap = states['defmap']
        defs = defmap[label]
        if defs:
            lastdef = defs[-1]
            return lastdef
        else:
            return self._find_def_from_top(states, label, loc=loc)

    def _stmt_index(self, defstmt, block, stop=-1):
        """Find the positional index of the statement at ``block``.

        Assumptions:
        - no two statements can point to the same object.
        """
        # Compare using id() as IR node equality is for semantic equivalence
        # opposed to direct equality (the location and scope are not considered
        # as part of the equality measure, this is important here).
        for i in range(len(block.body))[:stop]:
            if block.body[i] is defstmt:
                return i
        return len(block.body)


def _warn_about_uninitialized_variable(varname, loc):
    if config.ALWAYS_WARN_UNINIT_VAR:
        warnings.warn(
            errors.NumbaWarning(
                f"Detected uninitialized variable {varname}",
                loc=loc),
        )
